import os
import pandas as pd
from genewalk.cli import create_folder, default_base_folder
from genewalk.plot import GW_Plotter


def test_generate_plots():
    project = 'plot_test'
    project_folder = create_folder(default_base_folder, project)
    dGW = pd.read_csv(os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                   'test_results.csv'))
    figure_folder = create_folder(project_folder, 'figures')
    create_folder(figure_folder, 'barplots')
    GWp = GW_Plotter(figure_folder, dGW, 0.1)
    GWp.generate_plots()

    def file_exists(fname):
        return os.path.exists(os.path.join(figure_folder, fname))

    assert file_exists('moonlighters_x_go_con_y_frac_rel_go.png')
    assert file_exists('regulators_x_gene_con_y_frac_rel_go.png')
    assert file_exists('index.html')
    assert file_exists('barplots/barplot_BRAF_1097_x_mlog10global_padj_y_GO.png')
    assert file_exists('barplots/barplot_KRAS_6407_x_mlog10global_padj_y_GO.png')



import os
from genewalk.gene_lists import read_gene_list
from genewalk.nx_mg_assembler import UserNxMgAssembler
from genewalk.resources import ResourceManager
from .util import TEST_RESOURCES, TEST_BASE_FOLDER


rm = ResourceManager(TEST_BASE_FOLDER)
sif_genes = os.path.join(TEST_RESOURCES, 'test_sif.sif')
sif_annots = os.path.join(TEST_RESOURCES, 'test_sif_annot.sif')
sif_full = os.path.join(TEST_RESOURCES, 'test_sif_full.sif')
genes = read_gene_list(os.path.join(TEST_RESOURCES, 'hgnc_symbols.txt'),
                       id_type='hgnc_symbol', resource_manager=rm)


def test_gene_only_sif():
    mga = UserNxMgAssembler(genes, resource_manager=rm, filepath=sif_genes,
                            gwn_format='sif')
    gene_nodes = {'KRAS', 'BRAF', 'MAP2K2', 'MAPK1', 'PIK3CA', 'AKT1'}
    go_nodes = {'GO:0001934', 'GO:0005515', 'GO:0000186', 'GO:0000001',
                'GO:0032147', 'GO:0003924'}
    assert set(mga.graph.nodes()) == gene_nodes | go_nodes

    # Make sure we have GO node annotations as expected
    go_node = mga.graph.nodes['GO:0000186']
    assert go_node['GO'] == 'GO:0000186', go_node
    assert go_node['domain'] == 'biological_process', go_node
    assert go_node['name'] == 'activation of MAPKK activity', go_node

    # Make sure we have GO annotation edges
    assert ('BRAF', 'GO:0000186') in mga.graph.edges
    assert ('BRAF', 'GO:0005515') in mga.graph.edges

    # Make sure we have GO DAG edges
    assert ('GO:0000186', 'GO:0032147') in mga.graph.edges


def test_annot_sif():
    mga = UserNxMgAssembler(genes, resource_manager=rm, filepath=sif_annots,
                            gwn_format='sif_annot')
    gene_nodes = {'KRAS', 'BRAF', 'MAP2K2', 'MAPK1', 'PIK3CA', 'AKT1'}
    go_nodes = {'GO:0001934', 'GO:0005515', 'GO:0000186', 'GO:0000001',
                'GO:0032147', 'GO:0003924'}
    assert set(mga.graph.nodes()) == gene_nodes | go_nodes

    # Make sure we have GO node annotations as expected
    go_node = mga.graph.nodes['GO:0000186']
    assert go_node['GO'] == 'GO:0000186', go_node
    assert go_node['domain'] == 'biological_process', go_node
    assert go_node['name'] == 'activation of MAPKK activity', go_node

    # Make sure we have GO annotation edges
    assert ('BRAF', 'GO:0000186') in mga.graph.edges
    # This is in the test GO annotations file but not in the SIF so
    # it shouldn't be here
    assert ('BRAF', 'GO:0005515') not in mga.graph.edges

    # Make sure we have GO DAG edges
    assert ('GO:0000186', 'GO:0032147') in mga.graph.edges


def test_full_sif():
    mga = UserNxMgAssembler(genes, resource_manager=rm, filepath=sif_full,
                            gwn_format='sif_full')
    gene_nodes = {'KRAS', 'BRAF', 'MAP2K2', 'MAPK1', 'PIK3CA', 'AKT1'}
    # This time we only have the GO nodes that explicitly appear in the
    # SIF file
    go_nodes = {'GO:0005515', 'GO:0000186', 'GO:0001934'}
    assert set(mga.graph.nodes()) == gene_nodes | go_nodes

    # Make sure we have GO node annotations as expected
    go_node = mga.graph.nodes['GO:0000186']
    assert go_node['GO'] == 'GO:0000186', go_node
    assert go_node['domain'] == 'biological_process', go_node
    assert go_node['name'] == 'activation of MAPKK activity', go_node

    # Make sure we have GO annotation edges
    assert ('BRAF', 'GO:0000186') in mga.graph.edges
    # This is in the test GO annotations file but not in the SIF so
    # it shouldn't be here
    assert ('BRAF', 'GO:0005515') not in mga.graph.edges

    # Make sure we have GO DAG edges, and only ones that are in the
    # SIF, not ones in the GO ontology
    assert ('GO:0000186', 'GO:0032147') not in mga.graph.edges
    assert ('GO:0000186', 'GO:0001934') in mga.graph.edges


def test_custom_sif():
    custom_genes = read_gene_list(
        os.path.join(TEST_RESOURCES, 'custom_gene_list.txt'),
        id_type='custom', resource_manager=rm)
    sif = os.path.join(TEST_RESOURCES, 'test_sif_custom.sif')
    mga = UserNxMgAssembler(custom_genes, resource_manager=rm,
                            filepath=sif, gwn_format='sif_annot')
    # This is to make sure the network filtering works
    assert 'CUSTOM:XXXX' not in mga.graph
    assert 'CUSTOM:YYYY' not in mga.graph

    gene_nodes = {'CUSTOM:ABC', 'CUSTOM:XYZ', 'CUSTOM:ZZZ'}
    # We still have all the GO nodes that are in the test GO OBO
    # since we are in sif_annot mode here
    go_nodes = {'GO:0001934', 'GO:0005515', 'GO:0000186', 'GO:0000001',
                'GO:0032147', 'GO:0003924'}
    graph_nodes = set(mga.graph.nodes())
    assert graph_nodes == gene_nodes | go_nodes, graph_nodes

    # Make sure we have GO node annotations as expected
    go_node = mga.graph.nodes['GO:0000186']
    assert go_node['GO'] == 'GO:0000186', go_node
    assert go_node['domain'] == 'biological_process', go_node
    assert go_node['name'] == 'activation of MAPKK activity', go_node

    # Make sure we have gene-gene edges
    assert ('CUSTOM:ABC', 'CUSTOM:XYZ') in mga.graph.edges

    # Make sure we have GO annotation edges
    assert ('CUSTOM:ABC', 'GO:0000186') in mga.graph.edges

    # Make sure we have GO DAG edges
    assert ('GO:0000186', 'GO:0032147') in mga.graph.edges



import os
import pandas
import logging
from nose.tools import raises
from genewalk.cli import run_main
from .util import place_resource_files, TEST_RESOURCES, TEST_BASE_FOLDER

logger = logging.getLogger(__name__)


class ArgparseMock:
    def __init__(self, project, genes, id_type,
                 stage='all', base_folder=TEST_BASE_FOLDER,
                 network_source='pc', network_file=None, nproc=1,
                 nreps_graph=3, nreps_null=3, alpha_fdr=1.0,
                 dim_rep=8, save_dw=False, random_seed=None):
        self.project = project
        self.genes = genes
        self.id_type = id_type
        self.stage = stage
        self.base_folder = base_folder
        self.network_source = network_source
        self.network_file = network_file
        self.nproc = nproc
        self.nreps_graph = nreps_graph
        self.nreps_null = nreps_null
        self.alpha_fdr = alpha_fdr
        self.dim_rep = dim_rep
        self.save_dw = save_dw
        self.random_seed = random_seed


def test_default():
    project_name = 'test1'
    gene_list = os.path.join(TEST_RESOURCES, 'hgnc_symbols.txt')
    args = ArgparseMock(project_name, gene_list, 'hgnc_symbol')
    place_resource_files()
    run_main(args)

    assert os.path.exists(TEST_BASE_FOLDER)
    result_csv = os.path.join(TEST_BASE_FOLDER, project_name,
                              'genewalk_results.csv')
    assert os.path.exists(result_csv)
    df = pandas.read_csv(result_csv)
    assert 'MAP2K2' in set(df['hgnc_symbol']), df['hgnc_symbol']


def test_sif():
    project_name = 'test_sif'
    gene_list = os.path.join(TEST_RESOURCES, 'hgnc_symbols.txt')
    sif = os.path.join(TEST_RESOURCES, 'test_sif.sif')
    args = ArgparseMock(project_name, gene_list, 'hgnc_symbol',
                        network_source='sif', network_file=sif)
    place_resource_files()
    run_main(args)
    assert os.path.exists(TEST_BASE_FOLDER)
    result_csv = os.path.join(TEST_BASE_FOLDER, project_name,
                              'genewalk_results.csv')
    assert os.path.exists(result_csv)
    df = pandas.read_csv(result_csv)
    assert 'MAP2K2' in set(df['hgnc_symbol']), df['hgnc_symbol']
    assert 'GO:0005515' in set(df['go_id'])
    assert 'GO:0001934' in set(df['go_id'])
    assert 'biological process' in set(df['go_domain'])


def test_edge_list():
    project_name = 'test_edge_list'
    gene_list = os.path.join(TEST_RESOURCES, 'hgnc_symbols.txt')
    el = os.path.join(TEST_RESOURCES, 'test_edge_list.txt')
    args = ArgparseMock(project_name, gene_list, 'hgnc_symbol',
                        network_source='edge_list', network_file=el)
    place_resource_files()
    run_main(args)
    assert os.path.exists(TEST_BASE_FOLDER)
    result_csv = os.path.join(TEST_BASE_FOLDER, project_name,
                              'genewalk_results.csv')
    assert os.path.exists(result_csv)
    df = pandas.read_csv(result_csv)
    assert 'MAP2K2' in set(df['hgnc_symbol']), df['hgnc_symbol']
    assert 'GO:0005515' in set(df['go_id'])
    assert 'GO:0001934' in set(df['go_id'])
    assert 'biological process' in set(df['go_domain'])


def test_sif_annot():
    project_name = 'test_sif'
    gene_list = os.path.join(TEST_RESOURCES, 'hgnc_symbols.txt')
    sif = os.path.join(TEST_RESOURCES, 'test_sif_annot.sif')
    args = ArgparseMock(project_name, gene_list, 'hgnc_symbol',
                        network_source='sif_annot', network_file=sif)
    place_resource_files()
    run_main(args)
    assert os.path.exists(TEST_BASE_FOLDER)
    result_csv = os.path.join(TEST_BASE_FOLDER, project_name,
                              'genewalk_results.csv')
    assert os.path.exists(result_csv)
    df = pandas.read_csv(result_csv)
    assert 'MAP2K2' in set(df['hgnc_symbol']), df['hgnc_symbol']
    assert 'GO:0005515' in set(df['go_id'])
    # In this case we don't have this annotation in the SIF file
    assert 'GO:0001934' not in set(df['go_id'])
    assert 'biological process' in set(df['go_domain'])


def test_sif_full():
    project_name = 'test_sif'
    gene_list = os.path.join(TEST_RESOURCES, 'hgnc_symbols.txt')
    sif = os.path.join(TEST_RESOURCES, 'test_sif_full.sif')
    args = ArgparseMock(project_name, gene_list, 'hgnc_symbol',
                        network_source='sif_full', network_file=sif)
    place_resource_files()
    run_main(args)
    assert os.path.exists(TEST_BASE_FOLDER)
    result_csv = os.path.join(TEST_BASE_FOLDER, project_name,
                              'genewalk_results.csv')
    assert os.path.exists(result_csv)
    df = pandas.read_csv(result_csv)
    assert 'MAP2K2' in set(df['hgnc_symbol']), df['hgnc_symbol']
    assert 'GO:0005515' in set(df['go_id'])
    # In this case we don't have this annotation in the SIF file
    assert 'GO:0001934' not in set(df['go_id'])
    # It's particularly important here that we have e.g., GO domains
    # since this means that node attributes are added correctly for
    # GO terms mentioned in the SIF file
    assert 'biological process' in set(df['go_domain'])


def test_custom_genes():
    project_name = 'test_custom_genes'
    gene_list = os.path.join(TEST_RESOURCES, 'custom_gene_list.txt')
    sif = os.path.join(TEST_RESOURCES, 'test_sif_custom.sif')
    args = ArgparseMock(project_name, gene_list, 'custom',
                        network_source='sif_annot', network_file=sif)
    place_resource_files()
    run_main(args)
    assert os.path.exists(TEST_BASE_FOLDER)
    result_csv = os.path.join(TEST_BASE_FOLDER, project_name,
                              'genewalk_results.csv')
    assert os.path.exists(result_csv)
    df = pandas.read_csv(result_csv)
    assert 'CUSTOM:ABC' in set(df['custom']), df['custom']
    assert 'GO:0000186' in set(df['go_id'])
    # In this case we don't have this annotation in the SIF file
    assert 'GO:0001934' not in set(df['go_id'])
    assert 'biological process' in set(df['go_domain'])


@raises(ValueError)
def test_missing_network_file():
    project_name = 'test_missing_network_file'
    gene_list = os.path.join(TEST_RESOURCES, 'hgnc_symbols.txt')
    args = ArgparseMock(project_name, gene_list, 'hgnc_symbol',
                        network_source='sif')
    place_resource_files()
    run_main(args)



import os
import gzip
import shutil
import logging
import pandas
import urllib.request
from genewalk.gene_lists import GeneMapper

logger = logging.getLogger('genewalk.resources')


class ResourceManager(object):
    """Class to manage the download, caching and access of resource files."""
    def __init__(self, base_folder=None):
        self.base_folder = base_folder if base_folder else \
            os.path.join(os.path.expanduser('~'), 'genewalk')
        self.resource_folder = self._get_resource_folder()
        logger.info('Using %s as resource folder.' % self.resource_folder)

    def get_go_obo(self):
        fname = os.path.join(self.resource_folder, 'go.obo')
        if not os.path.exists(fname):
            url = 'https://snapshot.geneontology.org/ontology/go.obo'
            download_url(url, fname)
        return fname

    def get_goa_gaf(self):
        fname = os.path.join(self.resource_folder, 'goa_human.gaf')
        if not os.path.exists(fname):
            url_goa = ('https://geneontology.org/gene-associations/'
                       'goa_human.gaf.gz')
            download_gz(fname, url_goa)
        return fname

    def get_pc(self):
        fname_current = os.path.join(self.resource_folder,
                                     'PathwayCommons12.All.hgnc_current.sif')
        if not os.path.exists(fname_current):
            fname = os.path.join(self.resource_folder,
                                 'PathwayCommons12.All.hgnc.sif')
            if not os.path.exists(fname):
                url_pc = ('https://www.pathwaycommons.org/archives/PC2/v12/'
                          'PathwayCommons12.All.hgnc.sif.gz')
                download_gz(fname, url_pc)
            self._replace_outdated_hgnc_symbols(fname,fname_current)
        return fname_current

    def get_mgi_entrez(self):
        fname = os.path.join(self.resource_folder, 'MGI_EntrezGene.rpt')
        if not os.path.exists(fname):
            url = 'https://www.informatics.jax.org/downloads/reports/' \
                  'MGI_EntrezGene.rpt'
            download_url(url, fname)
        return fname

    def get_hgnc(self):
        fname = os.path.join(self.resource_folder, 'hgnc_entries.tsv')
        if not os.path.exists(fname):
            # Select relevant columns and parameters
            cols = ['gd_hgnc_id', 'gd_app_sym', 'gd_app_name', 'gd_prev_sym',
                    'gd_status', 'md_eg_id', 'md_prot_id', 'md_mgd_id',
                    'md_rgd_id', 'gd_pub_ensembl_id']
            statuses = ['Approved', 'Entry%20Withdrawn']
            params = {
                'hgnc_dbtag': 'on',
                'order_by': 'gd_app_sym_sort',
                'format': 'text',
                'submit': 'submit'
            }

            # Construct a download URL from the above parameters
            url = 'https://www.genenames.org/cgi-bin/download/custom?'
            url += '&'.join(['col=%s' % c for c in cols]) + '&'
            url += '&'.join(['status=%s' % s for s in statuses]) + '&'
            url += '&'.join(['%s=%s' % (k, v) for k, v in params.items()])

            # Save the download into a file
            download_url(url, fname)
        return fname

    def _get_resource_folder(self):
        resource_dir = os.path.join(self.base_folder, 'resources')
        if not os.path.isdir(resource_dir):
            try:
                os.makedirs(resource_dir)
            except Exception:
                logger.warning(resource_dir + ' already exists')
        return resource_dir

    def _replace_outdated_hgnc_symbols(self, pc_old, pc_current):
        gm = GeneMapper(self)
        logger.info('Replacing outdated HGNC symbols in %s and '
                    'saving as %s' % (pc_old, pc_current))
        pc = pandas.read_csv(pc_old, sep='\t', dtype=str, header=None)
        col_mapper = {0: 'source', 1: 'rel_type', 2: 'target'}
        pc = pc.rename(mapper=col_mapper, axis='columns')
        all_symbols = set(pc['source']).union(pc['target'])
        symbol_map = {}
        for sym in all_symbols:
            if not sym.startswith('CHEBI:'):
                hgnc_id = gm.get_current_hgnc_id(sym)
                if not hgnc_id:
                    continue
                elif isinstance(hgnc_id, list):
                    # outdated gene symbol is ambiguous: maps to multiple genes
                    continue
                latest_symbol = gm.get_hgnc_name(hgnc_id)
                if latest_symbol != sym:
                    symbol_map[sym] = latest_symbol
        if symbol_map:
            pc.replace(symbol_map,inplace=True)
        pc.to_csv(pc_current, sep='\t', header=False, index=False)
        os.remove(pc_old)

    def download_all(self):
        self.get_go_obo()
        self.get_goa_gaf()
        self.get_pc()
        self.get_hgnc()
        self.get_mgi_entrez()


def download_url(url, fname):
    logger.info('Downloading %s into %s' % (url, fname))
    urllib.request.urlretrieve(url, fname)


def download_gz(fname, url):
    logger.info('Downloading %s and extracting into %s' % (url, fname))
    gz_file = fname + '.gz'
    urllib.request.urlretrieve(url, gz_file)
    with gzip.open(gz_file, 'rb') as fin:
        with open(fname, 'wb') as fout:
            shutil.copyfileobj(fin, fout)


if __name__ == '__main__':
    # Download all the resources if this script is run directly
    ResourceManager().download_all()



import os.path
import logging
import pandas as pd
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
plt.rcParams['pdf.fonttype'] = 42
import seaborn as sns
import plotly.express as px

logger = logging.getLogger('genewalk.plot')


class GW_Plotter(object):
    """
    genewalk plotter object to visualize genewalk output data.

    Parameters
    ----------
    path : str
        Path to output figures (Default: project_folder/figures)
    dgw : pandas.dataFrame
        GeneWalk results data frame.
    alpha_fdr : float
        FDR significance level for global analyses and barplots.
        Default: 0.1 and must be < 1 otherwise set to default.
        Lower this value to increase the stringency of the
        regulator selection procedure.
    """
    def __init__(self, path, dgw, alpha_fdr):
        self.path = path
        if alpha_fdr == 1:  # default input
            self.alpha_fdr = 0.1
        else:
            self.alpha_fdr = alpha_fdr
        self.scatter_data = pd.DataFrame()
        self.stat = 'global_padj'  # for barplots
        self.ci_stat = 'cilow_'+self.stat
        self.dGW = dgw
        self.go_domains = set(self.dGW['go_domain'])
        if self.dGW.columns[0] in {'mgi_id', 'rgd_id', 'ensembl_id',
                                   'entrez_human', 'entrez_mouse',
                                   'custom'}:
            self.std_id = False
            self.id_type = self.dGW.columns[0]
        else:
            self.std_id = True
            self.id_type = self.dGW.columns[1]

        self.name_namespace = 'custom' if self.id_type == 'custom' \
            else 'hgnc_symbol'
        self.id_namespace = 'custom' if self.id_type == 'custom' \
            else 'hgnc_id'

    def generate_plots(self):
        """Wrapper that calls scatter and bar plot generating functions."""
        reg_html = self.scatterplot_regulators()
        moonlight_html = self.scatterplot_moonlighters()
        self.barplot_goanno()
        self.make_html([reg_html, moonlight_html])

    def scatterplot_regulators(self):
        """Scatter plot with fraction of (globally) relevant GO annotations
        as a function of gene connectivity (to other genes) for all input
        genes.

        Genes with symbols listed are regulator genes.
        See genewalk_scatterplots.csv for full data and
        genewalk_regulators.csv for regulator genes of interest.
        Visualization thresholds:
        T_frac: minimal fraction of relevant GO annotations, set to 0.5
        T_gcon: minimal gene connectivity (to other
        genes), set to 75th quartile of distribution
        """
        if self.scatter_data.empty:
            self._get_scatter_data()
        plot_title = 'Regulator genes'
        xvar = 'gene_con'
        yvar = 'frac_rel_go'
        xlab = 'Connections with other genes (per gene)'
        ylab = 'Fraction of relevant GO terms (per gene)'
        xmin = 0.45
        xmax = max(self.scatter_data[xvar])*1.2
        T_gcon = self.scatter_data[xvar].quantile(q=0.75)
        T_frac = 0.5

        # seaborn static plot
        sns.set(style="whitegrid")
        fig, ax = plt.subplots(figsize=(12, 12))  # inches
        g = sns.scatterplot(x=xvar, y=yvar, hue=yvar,
                            linewidth=0, alpha=0.5,
                            sizes=(40, 400),
                            data=self.scatter_data,
                            ax=ax, legend=False)
        plt.axvline(x=T_gcon, color='grey', linestyle='--')
        plt.axhline(y=T_frac, color='grey', linestyle='--')
        font_sz = 16
        plt.xlabel(xlab, size=font_sz)
        plt.ylabel(ylab, size=font_sz)
        plt.xlim([xmin, xmax])
        plt.xticks(size=font_sz)
        plt.yticks(size=font_sz)

        regulators = []
        dreg = self.scatter_data[self.scatter_data[xvar] >= T_gcon]
        dreg = dreg[dreg[yvar] >= T_frac]
        for r in dreg.index:
            gname = dreg[self.name_namespace][r]
            regulators.append(gname)
            x_txt = dreg[xvar][r]
            y_txt = dreg[yvar][r]+np.random.normal(0, 0.002)
            g.text(x_txt, y_txt, gname, size=6, horizontalalignment='center',
                   color='black', weight='light', fontstyle='italic')
        g.set(xscale="log")
        plt.title(plot_title, size=font_sz)
        filename = 'regulators_x_' + xvar + '_y_' + yvar
        plt.savefig(os.path.join(self.path, filename + '.pdf'),
                    bbox_inches="tight", transparent=True)
        plt.savefig(os.path.join(self.path, filename + '.png'),
                    bbox_inches="tight", transparent=True)

        # plotly interactive plot
        fig = px.scatter(self.scatter_data[~self.scatter_data[yvar].isna()],
                         x=xvar, y=yvar,
                         color=yvar, size='rel_go',
                         hover_name=self.name_namespace,
                         hover_data=[self.name_namespace, self.id_type],
                         title=plot_title, labels={xvar: xlab, yvar: ylab},
                         log_x=True, range_x=[xmin, xmax])
        fig.add_shape(type='rect', x0=T_gcon, y0=T_frac, x1=xmax, y1=1,
                      fillcolor="LightSkyBlue", opacity=0.2,
                      layer="below", line_width=0)
        fig.add_shape(type='line', x0=xmin, y0=T_frac, x1=xmax, y1=T_frac,
                      line=dict(color='grey', dash='dash'))
        fig.add_shape(type='line', x0=T_gcon, y0=0, x1=T_gcon, y1=1,
                      line=dict(color='grey', dash='dash'))
        plotly_html = fig.to_html(full_html=False)
        fig.write_html(os.path.join(self.path, filename + '.html'))
        logger.info('%s plotted in %s...' % (plot_title, filename))

        df = pd.DataFrame(sorted(regulators), columns=['gw_regulators'])
        filename = 'genewalk_regulators.csv'
        df.to_csv(os.path.join(self.path, filename), index=False)
        logger.info('%s listed in %s...' % (plot_title, filename))
        return plotly_html

    def scatterplot_moonlighters(self):
        """Scatter plot with fraction of (globally) relevant GO annotations
        as a function of number of GO annotations for all input genes.

        Genes with symbols listed are moonlighting genes.
        See genewalk_scatterplots.csv for full data and
        genewalk_moonlighters.csv for moonlighting genes of interest.
        Visualization thresholds:
        T_frac: maximal fraction of relevant GO annotations, set to 0.5
        T_gocon: minimal number of GO annotations per gene,
        set to max of 30 and 75th quartile of distribution.
        """
        if self.scatter_data.empty:
            self._get_scatter_data()
        plot_title = 'Moonlighting genes'
        xvar = 'go_con'
        yvar = 'frac_rel_go'
        xlab = 'Number of GO annotations (per gene)'
        ylab = 'Fraction of relevant GO terms (per gene)'
        xmin = 0.45
        xmax = max(self.scatter_data[xvar])*1.2
        T_gocon = max(30, self.scatter_data[xvar].quantile(q=0.75))
        T_frac = 0.5

        #seaborn static plot
        sns.set(style="whitegrid")
        fig, ax = plt.subplots(figsize=(12, 12))  # inches
        g = sns.scatterplot(x=xvar, y=yvar, hue=yvar,
                            linewidth=0, alpha=0.5,
                            sizes=(40, 400),
                            data=self.scatter_data,
                            ax=ax, legend=False)
        plt.axvline(x=T_gocon, color='grey', linestyle='--')
        plt.axhline(y=T_frac, color='grey', linestyle='--')
        font_sz=16
        plt.xlabel(xlab, size=font_sz)
        plt.ylabel(ylab, size=font_sz)
        plt.xlim([xmin, xmax])
        plt.xticks(size=font_sz)
        plt.yticks(size=font_sz)

        moonlighters = []
        dmoon = self.scatter_data[self.scatter_data[xvar] >= T_gocon]
        dmoon = dmoon[(dmoon[yvar] < T_frac) & (dmoon[yvar] > 0)]
        for m in dmoon.index:
            gname = dmoon[self.name_namespace][m]
            moonlighters.append(gname)
            x_txt = dmoon[xvar][m]
            y_txt = dmoon[yvar][m]
            g.text(x_txt, y_txt, gname, size=6, horizontalalignment='center',
                   color='black', weight='light',
                   fontstyle='italic')
        g.set(xscale="log")
        plt.title(plot_title, size=font_sz)
        filename = 'moonlighters_x_' + xvar + '_y_' + yvar
        plt.savefig(os.path.join(self.path, filename + '.pdf'),
                    bbox_inches="tight", transparent=True)
        plt.savefig(os.path.join(self.path, filename + '.png'),
                    bbox_inches="tight", transparent=True)

        ### plotly interactive plot
        fig = px.scatter(self.scatter_data[~self.scatter_data[yvar].isna()],
                         x=xvar, y=yvar,
                         color=yvar, size='gene_con',
                         hover_name=self.name_namespace,
                         hover_data=[self.name_namespace, self.id_type],
                         title=plot_title, labels={xvar: xlab, yvar: ylab},
                         log_x=True, range_x=[xmin, xmax])
        fig.add_shape(type='rect', x0=T_gocon, y0=0, x1=xmax, y1=T_frac,
                      fillcolor="LightSkyBlue", opacity=0.2,
                      layer="below", line_width=0)
        fig.add_shape(type='line', x0=xmin, y0=T_frac, x1=xmax, y1=T_frac,
                      line=dict(color='grey', dash='dash'))
        fig.add_shape(type='line', x0=T_gocon, y0=0, x1=T_gocon, y1=1,
                      line=dict(color='grey', dash='dash'))
        fig.write_html(os.path.join(self.path, filename + '.html'))
        plotly_html = fig.to_html(full_html=False)
        logger.info('%s plotted in %s...' % (plot_title, filename))

        df = pd.DataFrame(sorted(moonlighters), columns=['gw_moonlighter'])
        filename = 'genewalk_moonlighters.csv'
        df.to_csv(os.path.join(self.path, filename), index=False)
        logger.info('%s listed in %s...' % (plot_title, filename))
        return plotly_html

    def barplot_goanno(self):
        """Visualize statistical significances of all GO annotations in a barplot
        for a given gene of interest.
        """
        self.dGW['mlog10padj'] = -np.log10(self.dGW[self.stat])
        self.dGW['mlog10padj_err'] = - np.log10(self.dGW[self.ci_stat]) \
                                     - self.dGW['mlog10padj']
        for gid in self.dGW[self.id_type].unique():
            df = self.dGW[self.dGW[self.id_type] == gid]
            gsymbol = self.scatter_data[self.name_namespace][gid]
            self._barplot(df, gid, gsymbol, dom='GO')
            #Barplots separated by go domain
            #for go_dom in self.go_domains:
            #    self._barplot(df[df['go_domain']==go_dom], gid, gsymbol,dom=go_dom)

    def _get_scatter_data(self):
        """
        Data processing function for scatter plots.
        """
        scd = dict()
        if not self.std_id:
            if self.id_type == 'custom':
                scat_cols = [self.id_type]
            else:
                scat_cols = [self.id_type, self.name_namespace,
                             self.id_namespace]
            scat_cols += ['con', 'go_con', 'gene_con', 'rel_go',
                          'frac_rel_go']
        else:
            scat_cols = [self.id_type, self.name_namespace, 'con', 'go_con',
                         'gene_con', 'rel_go', 'frac_rel_go']
        for gid in self.dGW[self.id_type].unique():
            df = self.dGW[ self.dGW[self.id_type] == gid ]
            gname = df[self.name_namespace].unique()[0]
            con = df['ncon_gene'].unique()[0]
            if pd.isna(df['go_id'].unique()[0]):  # no GO annotations
                gocon = np.nan
                genecon = con
                relgo = 0
                fracrelgo = 0
            else:
                gocon = len(df)
                genecon = con - gocon
                relgo = min(len(df[df['global_padj'] < self.alpha_fdr]),
                            len(df[df['gene_padj'] < self.alpha_fdr]))
                fracrelgo = round(float(relgo)/gocon, 3)

            if self.std_id:
                scd[gid] = [gid, gname, con, gocon,
                            genecon, relgo, fracrelgo]
            else:
                hid = str(df[self.id_namespace].unique()[0])
                if self.id_type == 'custom':
                    scd[gid] = [gid]
                else:
                    scd[gid] = [gid, gname, hid]
                scd[gid] += [con, gocon, genecon, relgo, fracrelgo]
        self.scatter_data = pd.DataFrame.from_dict(scd, orient='index',
                                                   columns=scat_cols)
        self.scatter_data.sort_values(by=['gene_con', 'frac_rel_go', 'rel_go'],
                                      ascending=[False, False, False],
                                      inplace=True)
        filename = 'genewalk_scatterplots.csv'
        self.scatter_data.to_csv(os.path.join(self.path, filename),
                                 index=False)
        logger.info('Scatter plot data output to %s...' % filename)
        for c in ['go_con', 'gene_con']:  # for log scale plotting: 0 -> 0.5
            self.scatter_data[c].where(self.scatter_data[c] > 0, 0.5,
                                       inplace=True)

    def _barplot(self, dplot, gene_id, gsymbol, dom):
        """
        Bar plot figure generating function.
        """
        gocon = self.scatter_data['go_con'][gene_id]
        gene_id = str(gene_id)
        if gocon >= 1:
            dplot = dplot.sort_values(by=['mlog10padj', 'sim',
                                          'global_padj', 'gene_padj'],
                                      ascending=[False, False,
                                                 True, True])
            sns.set(style="whitegrid", color_codes=True)
            f, ax = plt.subplots(figsize=(2, 0.25*len(dplot)))
            ax = sns.barplot(x='mlog10padj', y='go_name',
                             xerr=dplot['mlog10padj_err'],
                             data=dplot, color="b",
                             error_kw=dict(elinewidth=1,
                                           ecolor=[0.7, 0.7, 0.7], capsize=1))
            plt.axvline(x=-np.log10(self.alpha_fdr), color='r', linestyle='--')
            font_sz = 12
            max_x_data = max(dplot['mlog10padj']+dplot['mlog10padj_err'])+0.3
            plt.xticks(range(round(max_x_data)+1), size=font_sz)
            plt.xlim(0, max(round(max_x_data, 1), 1))
            plt.xlabel('-log10('+self.stat+')', size=font_sz)
            ax.yaxis.set_label_position('right')
            plt.ylabel(dom+' annotations', size=0.8*font_sz,
                       rotation=270, labelpad=15)
            plt.title(gsymbol, size=font_sz)
            filename = 'barplot_' + gsymbol + '_' + gene_id + '_x_mlog10' + \
                self.stat + '_y_' + dom.replace(' ', '') + '.'
            #plt.savefig(os.path.join(self.path, 'barplots', filename+'pdf'),
            #            bbox_inches="tight", transparent=True)
            #pdf file size: >300kb, so memorywise not advisable to make
            #pdf barplots in case of long input gene lists.
            #Save as png (10kb) instead:
            plt.savefig(os.path.join(self.path, 'barplots', filename+'png'),
                        bbox_inches="tight", transparent=True)
            logger.info('Barplot generated for %s in %s...' %
                        (gene_id, filename))
        else:
            logger.warning('No results for gene id %s: \
                            could not produce barplot' % gene_id)

    def make_html(self, global_htmls):
        """
        Generates index.html, the html file that shows all the visualizations.
        """
        template_path = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), 'results_template.html')
        with open(template_path, 'r') as fh:
            template = fh.read()

        template = template.replace('{{ GLOBAL_RESULTS }}',
                                    '\n'.join(global_htmls))

        genes = iter(self.dGW[self.id_type].unique())
        gene_results_html = ""
        while True:
            gene_results_html += '<div class="row">'
            try:
                for _ in range(4):
                    gid = next(genes)
                    gsymbol = self.scatter_data[self.name_namespace][gid]
                    img_path = os.path.join('barplots',
                               ('barplot_%s_%s_x_mlog10%s_y_GO.png' %
                                (gsymbol, gid, self.stat)))
                    # Skip any genes for which results were not generated
                    if not os.path.exists(os.path.join(self.path, img_path)):
                        continue
                    # We only make proper links for non-custom gene symbols
                    if self.id_type != 'custom':
                        identifiers_href = 'https://identifiers.org/hgnc.symbol:%s' % gsymbol
                    else:
                        identifiers_href = '#'
                    gene_results_html += """
                    <div class="col-md-3">
                        <div class="thumbnail">
                            <a href="{img_path}">
                                <img src='{img_path}' style="width:100%">
                            </a>
                            <div class="caption">
                                <a href="{identifiers_href}">{symbol} ({gid})</a>
                            </div>
                        </div>
                    </div>
                    """.format(symbol=gsymbol, gid=gid, img_path=img_path,
                               identifiers_href=identifiers_href)
            except StopIteration:
                break
            finally:
                gene_results_html += '</div>'

        template = template.replace('{{ GENE_RESULTS }}', gene_results_html)
        output_html = os.path.join(self.path, 'index.html')
        with open(output_html, 'w') as fh:
            fh.write(template)
        logger.info('index.html file generated with interactive visualizations ' 
                    'of all GeneWalk results...')



import os
import glob
import shutil
import logging
from genewalk.cli import default_base_folder

logger = logging.getLogger(__name__)

TEST_MODULE = os.path.dirname(os.path.abspath(__file__))
TEST_RESOURCES = os.path.join(TEST_MODULE, 'resources')
TEST_BASE_FOLDER = os.path.join(default_base_folder, '.test')


def place_resource_files():
    test_resource_files = glob.glob(os.path.join(TEST_RESOURCES, '*'))
    test_resource_folder = \
        os.path.join(TEST_BASE_FOLDER, 'resources')
    os.makedirs(test_resource_folder, exist_ok=True)
    for test_file in test_resource_files:
        logger.debug('Copying %s into %s' % (test_file, test_resource_folder))
        shutil.copy(test_file,
                    os.path.join(test_resource_folder,
                                 os.path.basename(test_file)))



import re
import csv
import logging


logger = logging.getLogger('genewalk.gene_lists')


def read_gene_list(fname, id_type, resource_manager):
    """Return references for genes from a file with the given ID type.

    Parameters
    ----------
    fname : str
        The name of the file containing the list of genes. Each line of the
        file corresponds to a single gene.
    id_type : str
        The type of identifier contained in each line of the gene list file.
        Possible values are: hgnc_symbol, hgnc_id, ensembl_id, mgi_id.
    resource_manager : genewalk.resources.ResourceManager
        ResourceManager object, used to obtain entrez-mgi mappings if
        necessary.

    Returns
    -------
    dict
        A dictionary of references with keys including HGNCSYMBOL, HGNC, UP,
        and if id_type is mgi_id, MGI, with values corresponding to the
        identifiers of the provided list of genes.
    """
    gene_mapper = GeneMapper(resource_manager)
    with open(fname, 'r') as fh:
        # This is to make the list unique while preserving
        # the original order as much as possible
        unique_lines = []
        for line in fh.readlines():
            line = line.strip()
            if line not in unique_lines:
                unique_lines.append(line)
    if id_type == 'hgnc_symbol':
        refs = map_hgnc_symbols(unique_lines, gene_mapper)
    elif id_type == 'hgnc_id':
        refs = map_hgnc_ids(unique_lines, gene_mapper)
    elif id_type == 'ensembl_id':
        refs = map_ensembl_ids(unique_lines, gene_mapper)
    elif id_type == 'mgi_id':
        refs = map_mgi_ids(unique_lines, gene_mapper)
    elif id_type == 'rgd_id':
        refs = map_rgd_ids(unique_lines, gene_mapper)
    elif id_type == 'entrez_human':
        refs = map_entrez_human(unique_lines, gene_mapper)
    elif id_type == 'entrez_mouse':
        refs = map_entrez_mouse(unique_lines, gene_mapper)
    elif id_type == 'custom':
        refs = [{'ID': c} for c in unique_lines]
    else:
        raise ValueError('Unknown id_type: %s' % id_type)
    if not refs:
        raise ValueError('None of the IDs in %s could be mapped. It is '
                         'likely that the file uses an ID type or format '
                         'that GeneWalk cannot interpret.' % fname)
    return refs


def map_hgnc_symbols(hgnc_symbols, gene_mapper):
    """Return references based on a list of HGNC symbols."""
    refs = []
    for hgnc_symbol in hgnc_symbols:
        ref = {'HGNC_SYMBOL': hgnc_symbol, 'HGNC': None, 'UP': None}
        hgnc_id = gene_mapper.get_current_hgnc_id(hgnc_symbol)
        if not hgnc_id:
            logger.warning('Could not get HGNC ID for symbol %s' % hgnc_symbol)
            continue
        elif isinstance(hgnc_id, list):
            logger.warning('More than one current HGNC ID for outdated '
                           'symbol %s' % hgnc_symbol)
            continue
        ref['HGNC'] = hgnc_id
        uniprot_id = gene_mapper.get_uniprot_id(hgnc_id)
        if not uniprot_id:
            logger.warning('Could not get UniProt ID for symbol %s' %
                           hgnc_symbol)
            continue
        ref['UP'] = uniprot_id
        refs.append(ref)
    return refs


def map_hgnc_ids(hgnc_ids, gene_mapper):
    """Return references based on a list of HGNC IDs."""
    refs = []
    for hgnc_id in hgnc_ids:
        if hgnc_id.startswith('HGNC:'):
            hgnc_id = hgnc_id[5:]
        hgnc_ref = _refs_from_hgnc_id(hgnc_id, gene_mapper)
        if hgnc_ref is None:
            continue
        refs.append(hgnc_ref)
    return refs


def _refs_from_hgnc_id(hgnc_id, gene_mapper):
    ref = {'HGNC_SYMBOL': None, 'HGNC': hgnc_id, 'UP': None}
    hgnc_name = gene_mapper.get_hgnc_name(hgnc_id)
    if not hgnc_name:
        logger.warning('Could not get HGNC name for ID %s' %
                       hgnc_id)
        return None
    ref['HGNC_SYMBOL'] = hgnc_name
    uniprot_id = gene_mapper.get_uniprot_id(hgnc_id)
    if not uniprot_id:
        logger.warning('Could not get UniProt ID for HGNC ID %s' %
                       hgnc_id)
        return None
    ref['UP'] = uniprot_id
    return ref


def map_mgi_ids(mgi_ids, gene_mapper):
    """Return references based on a list of MGI IDs."""
    refs = []
    for mgi_id in mgi_ids:
        if mgi_id.startswith('MGI:'):
            mgi_id = mgi_id[4:]
        mgi_ref = _refs_from_mgi_id(mgi_id, gene_mapper)
        if mgi_ref is None:
            continue
        refs.append(mgi_ref)
    return refs


def _refs_from_mgi_id(mgi_id, gene_mapper):
    ref = {'MGI': mgi_id}
    hgnc_id = gene_mapper.get_hgnc_from_mgi(mgi_id)
    if hgnc_id is None:
        logger.warning('Could not get HGNC ID for MGI ID %s' %
                       mgi_id)
        return None
    hgnc_ref = _refs_from_hgnc_id(hgnc_id, gene_mapper)
    if hgnc_ref is None:
        return None
    ref.update(hgnc_ref)
    return ref


def map_rgd_ids(rgd_ids, gene_mapper):
    """Return references based on a list of RGD IDs."""
    refs = []
    for rgd_id in rgd_ids:
        if rgd_id.startswith('RGD:'):
            rgd_id = rgd_id[4:]
        rgd_ref = _refs_from_rgd_id(rgd_id, gene_mapper)
        if rgd_ref is None:
            continue
        refs.append(rgd_ref)
    return refs


def _refs_from_rgd_id(rgd_id, gene_mapper):
    ref = {'RGD': rgd_id}
    hgnc_id = gene_mapper.get_hgnc_from_rgd(rgd_id)
    if hgnc_id is None:
        logger.warning('Could not get HGNC ID for RGD ID %s' %
                       rgd_id)
        return None
    hgnc_ref = _refs_from_hgnc_id(hgnc_id, gene_mapper)
    if hgnc_ref is None:
        return None
    ref.update(hgnc_ref)
    return ref


def map_ensembl_ids(ensembl_ids, gene_mapper):
    """Return references based on a list of Ensembl IDs."""
    refs = []
    for ensembl_id in ensembl_ids:
        ref = {'ENSEMBL': ensembl_id}
        ensembl_id = ensembl_id.split('.', maxsplit=1)[0]
        hgnc_id = gene_mapper.get_hgnc_from_ensembl(ensembl_id)
        if not hgnc_id:
            logger.warning('Could not get HGNC ID for ENSEMBL ID %s' %
                           ensembl_id)
            continue
        hgnc_ref = _refs_from_hgnc_id(hgnc_id, gene_mapper)
        if hgnc_ref is None:
            continue
        ref.update(hgnc_ref)
        refs.append(ref)
    return refs


def map_entrez_human(entrez_ids, gene_mapper):
    """Return references based on human Entrez gene IDs."""
    refs = []
    for entrez_id in entrez_ids:
        ref = {'EGID': entrez_id}
        hgnc_id = gene_mapper.get_hgnc_from_entrez(entrez_id)
        if hgnc_id is None:
            logger.warning("Could not find HGNC ID for Entrez ID %s" %
                           entrez_id)
            continue
        hgnc_ref = _refs_from_hgnc_id(hgnc_id, gene_mapper)
        if hgnc_ref is None:
            continue
        ref.update(hgnc_ref)
        refs.append(ref)
    return refs


def map_entrez_mouse(entrez_ids, gene_mapper):
    """Return references based on mouse Entrez gene IDs."""
    # Get the entrez file path from the resource manager
    refs = []
    for entrez_id in entrez_ids:
        mgi_id = gene_mapper.entrez_to_mgi.get(entrez_id)
        if not mgi_id:
            logger.warning("Could not find an MGI mapping for Entrez ID %s"
                           % entrez_id)
            continue
        ref = {'EGID': entrez_id, 'MGI': mgi_id}
        mgi_refs = _refs_from_mgi_id(mgi_id, gene_mapper)
        if mgi_refs is None:
            continue
        ref.update(mgi_refs)
        refs.append(ref)
    return refs


class GeneMapper:
    def __init__(self, resource_manager):
        self.resource_manager = resource_manager
        self.hgnc_file = self.resource_manager.get_hgnc()
        self.mgi_entrez_file = self.resource_manager.get_mgi_entrez()

        # Process the MGI-Entrez mapping file
        self.entrez_to_mgi = {}
        with open(self.mgi_entrez_file, 'r') as fh:
            csvreader = csv.reader(fh, delimiter='\t')
            for row in csvreader:
                # Remove "MGI:" prefix
                mgi = row[0][4:]
                entrez = row[8]
                self.entrez_to_mgi[entrez] = mgi

        self.hgnc_id_to_name = {}
        self.hgnc_name_to_id = {}
        self.hgnc_withdrawn_to_new = {}
        self.hgnc_to_uniprot = {}
        self.mgi_to_hgnc = {}
        self.rgd_to_hgnc = {}
        self.entrez_to_hgnc = {}
        self.ensembl_to_hgnc = {}
        self.prev_sym_map = {}

        with open(self.hgnc_file, 'r', encoding='utf-8') as fh:
            csvreader = csv.reader(fh, delimiter='\t')
            # Skip the header
            next(csvreader)
            for row in csvreader:
                hgnc_id, hgnc_name, description, prev_sym_entry, hgnc_status,\
                    entrez_id, uniprot_id, mgi_id, rgd_id, ensembl_id = row
                hgnc_id = hgnc_id[5:]
                if hgnc_status in {'Approved', 'Entry Withdrawn'}:
                    self.hgnc_id_to_name[hgnc_id] = hgnc_name
                    # Note that withdrawn entries don't overlap with approved
                    # entries at this point so it's safe to add mappings for
                    # withdrawn names
                    self.hgnc_name_to_id[hgnc_name] = hgnc_id
                elif hgnc_status == 'Symbol Withdrawn':
                    m = re.match(r'symbol withdrawn, see \[HGNC:(?: ?)(\d+)\]',
                                 description)
                    new_id = m.groups()[0]
                    self.hgnc_withdrawn_to_new[hgnc_id] = new_id
                # Uniprot
                if uniprot_id:
                    self.hgnc_to_uniprot[hgnc_id] = uniprot_id
                # Entrez
                if entrez_id:
                    self.entrez_to_hgnc[entrez_id] = hgnc_id
                # Mouse
                if mgi_id:
                    mgi_ids = mgi_id.split(', ')
                    for mgi_id in mgi_ids:
                        if mgi_id.startswith('MGI:'):
                            mgi_id = mgi_id[4:]
                        self.mgi_to_hgnc[mgi_id] = hgnc_id
                # Rat
                if rgd_id:
                    rgd_ids = rgd_id.split(', ')
                    for rgd_id in rgd_ids:
                        if rgd_id.startswith('RGD:'):
                            rgd_id = rgd_id[4:]
                        self.rgd_to_hgnc[rgd_id] = hgnc_id
                # Previous symbols
                if prev_sym_entry:
                    prev_syms = prev_sym_entry.split(', ')
                    for prev_sym in prev_syms:
                        # If we already mapped this previous symbol
                        # to another ID
                        if prev_sym in self.prev_sym_map:
                            # If we already have a list here, we just extend it
                            if isinstance(self.prev_sym_map[prev_sym], list):
                                self.prev_sym_map[prev_sym].append(hgnc_id)
                            # Otherwise we create a list and start it with the
                            # two IDs we know the symbol is mapped to
                            else:
                                self.prev_sym_map[prev_sym] = \
                                    [self.prev_sym_map[prev_sym], hgnc_id]
                        # Otherwise we just make a string entry here
                        else:
                            self.prev_sym_map[prev_sym] = hgnc_id
                # Ensembl IDs
                if ensembl_id:
                    self.ensembl_to_hgnc[ensembl_id] = hgnc_id
            for old_id, new_id in self.hgnc_withdrawn_to_new.items():
                self.hgnc_id_to_name[old_id] = self.hgnc_id_to_name[new_id]

    def get_hgnc_name(self, hgnc_id):
        """Return the HGNC symbol corresponding to the given HGNC ID.

        Parameters
        ----------
        hgnc_id : str
            The HGNC ID to be converted.

        Returns
        -------
        hgnc_name : str
            The HGNC symbol corresponding to the given HGNC ID.
        """
        hgnc_name = self.hgnc_id_to_name.get(hgnc_id)
        return hgnc_name

    def get_hgnc_id(self, hgnc_name):
        """Return the HGNC ID corresponding to the given HGNC symbol.

        Parameters
        ----------
        hgnc_name : str
            The HGNC symbol to be converted. Example: BRAF

        Returns
        -------
        hgnc_id : str
            The HGNC ID corresponding to the given HGNC symbol.
        """
        return self.hgnc_name_to_id.get(hgnc_name)

    def get_current_hgnc_id(self, hgnc_name):
        """Return HGNC ID(s) corresponding to a current or outdated HGNC symbol.

        Parameters
        ----------
        hgnc_name : str
            The HGNC symbol to be converted, possibly an outdated symbol.

        Returns
        -------
        str or list of str or None
            If there is a single HGNC ID corresponding to the given current or
            outdated HGNC symbol, that ID is returned as a string. If the symbol
            is outdated and maps to multiple current IDs, a list of these
            IDs is returned. If the given name doesn't correspond to either
            a current or an outdated HGNC symbol, None is returned.
        """
        hgnc_id = self.get_hgnc_id(hgnc_name)
        if hgnc_id:
            return hgnc_id
        hgnc_id = self.prev_sym_map.get(hgnc_name)
        return hgnc_id

    def get_uniprot_id(self, hgnc_id):
        """Return the UniProt ID corresponding to the given HGNC ID.

        Parameters
        ----------
        hgnc_id : str
            The HGNC ID to be converted. Note that the HGNC ID is a number that is
            passed as a string. It is not the same as the HGNC gene symbol.

        Returns
        -------
        uniprot_id : str
            The UniProt ID corresponding to the given HGNC ID.
        """
        uniprot_id = self.hgnc_to_uniprot.get(hgnc_id)
        # The lookup can yield an empty string. Instead return None.
        if not uniprot_id:
            return None
        return uniprot_id

    def get_hgnc_from_entrez(self, entrez_id):
        """Return the HGNC ID corresponding to the given Entrez ID.

        Parameters
        ----------
        entrez_id : str
            The Entrez ID to be converted, a number passed as a string.

        Returns
        -------
        hgnc_id : str
            The HGNC ID corresponding to the given Entrez ID.
        """
        hgnc_id = self.entrez_to_hgnc.get(entrez_id)
        return hgnc_id

    def get_hgnc_from_ensembl(self, ensembl_id):
        """Return the HGNC ID corresponding to the given Ensembl ID.

        Parameters
        ----------
        ensembl_id : str
            The Ensembl ID to be converted, a number passed as a string.

        Returns
        -------
        hgnc_id : str
            The HGNC ID corresponding to the given Ensembl ID.
        """
        return self.ensembl_to_hgnc.get(ensembl_id)

    def get_hgnc_from_mgi(self, mgi_id):
        """Return the HGNC ID corresponding to the given MGI mouse gene ID.

        Parameters
        ----------
        mgi_id : str
            The MGI ID to be converted. Example: "2444934"

        Returns
        -------
        hgnc_id : str
            The HGNC ID corresponding to the given MGI ID.
        """
        if mgi_id.startswith('MGI:'):
            mgi_id = mgi_id[4:]
        return self.mgi_to_hgnc.get(mgi_id)

    def get_hgnc_from_rgd(self, rgd_id):
        """Return the HGNC ID corresponding to the given RGD rat gene ID.

        Parameters
        ----------
        rgd_id : str
            The RGD ID to be converted. Example: "1564928"

        Returns
        -------
        hgnc_id : str
            The HGNC ID corresponding to the given RGD ID.
        """
        if rgd_id.startswith('RGD:'):
            rgd_id = rgd_id[4:]
        return self.rgd_to_hgnc.get(rgd_id)



import os
from nose.tools import raises
from genewalk.gene_lists import *
from genewalk.cli import default_base_folder
from genewalk.resources import ResourceManager
from .util import TEST_RESOURCES

rm = ResourceManager()
gm = GeneMapper(rm)


def test_map_lists():
    refs = map_hgnc_symbols(['BRAF', 'KRAS'], gm)
    assert refs[0]['HGNC'] == '1097', refs
    assert refs[0]['UP'] == 'P15056', refs
    assert refs[0]['HGNC_SYMBOL'] == 'BRAF', refs
    assert refs[1]['HGNC'] == '6407', refs
    assert refs[1]['UP'] == 'P01116', refs
    assert refs[1]['HGNC_SYMBOL'] == 'KRAS', refs

    refs = map_hgnc_ids(['1097', '6407'], gm)
    assert refs[0]['HGNC'] == '1097', refs
    assert refs[0]['UP'] == 'P15056', refs
    assert refs[0]['HGNC_SYMBOL'] == 'BRAF', refs
    assert refs[1]['HGNC'] == '6407', refs
    assert refs[1]['UP'] == 'P01116', refs
    assert refs[1]['HGNC_SYMBOL'] == 'KRAS', refs

    refs = map_mgi_ids(['MGI:892970'], gm)
    assert refs[0]['HGNC'] == '6817', refs
    assert refs[0]['HGNC_SYMBOL'] == 'MAL', refs
    assert refs[0]['UP'] == 'P21145', refs

    refs = map_ensembl_ids(['ENSG00000157764'], gm)
    assert refs[0]['HGNC'] == '1097', refs
    assert refs[0]['UP'] == 'P15056', refs
    assert refs[0]['HGNC_SYMBOL'] == 'BRAF', refs

    refs = map_ensembl_ids(['ENSG00000157764.9'], gm)
    assert refs[0]['HGNC'] == '1097', refs
    assert refs[0]['UP'] == 'P15056', refs
    assert refs[0]['HGNC_SYMBOL'] == 'BRAF', refs


def test_read_gene_list():
    with open('test_gene_list.txt', 'w') as fh:
        fh.write('HGNC:1097')
    refs = read_gene_list('test_gene_list.txt', 'hgnc_id', rm)
    assert len(refs) == 1


@raises(ValueError)
def test_read_gene_list_bad():
    with open('test_gene_list.txt', 'w') as fh:
        fh.write('HGNC:1097')
    refs = read_gene_list('test_gene_list.txt', 'ensembl_id', rm)
    assert len(refs) == 1


def test_read_gene_list_entrez_human():
    with open('test_gene_list_entrez_human.txt', 'w') as fh:
        fh.write('2597')
    refs = read_gene_list('test_gene_list_entrez_human.txt',
                          'entrez_human', rm)
    assert refs[0]['HGNC_SYMBOL'] == 'GAPDH'
    assert len(refs) == 1


def test_read_gene_list_entrez_mouse():
    rm = ResourceManager(base_folder=default_base_folder)
    with open('test_gene_list_entrez_mouse.txt', 'w') as fh:
        fh.write('14433')
    refs = read_gene_list('test_gene_list_entrez_mouse.txt',
                          'entrez_mouse', rm)
    assert len(refs) == 1
    assert refs[0]['MGI'] == '95640'
    assert refs[0]['HGNC_SYMBOL'] == 'GAPDH'


def test_read_gene_list_rgd():
    rm = ResourceManager(base_folder=default_base_folder)
    with open('test_gene_list_rgd.txt', 'w') as fh:
        fh.write('2561\n')
        fh.write('RGD:69323')
    refs = read_gene_list('test_gene_list_rgd.txt',
                          'rgd_id', rm)
    assert len(refs) == 2, refs
    assert refs[0]['RGD'] == '2561'
    assert refs[0]['HGNC_SYMBOL'] == 'ERBB2'
    assert refs[1]['RGD'] == '69323'
    assert refs[1]['HGNC_SYMBOL'] == 'ERBB3'


def test_read_custom_list():
    rm = ResourceManager(base_folder=default_base_folder)
    gene_list_file = os.path.join(TEST_RESOURCES, 'custom_gene_list.txt')
    refs = read_gene_list(gene_list_file, 'custom', rm)
    assert len(refs) == 3, refs
    assert refs[0] == {'ID': 'CUSTOM:ABC'}, refs



import logging
import pandas as pd
import numpy as np
import networkx as nx
from statsmodels.stats.multitest import fdrcorrection
from scipy.stats import gmean, gstd

logger = logging.getLogger('genewalk.perform_statistics')


class GeneWalk(object):
    """GeneWalk object that generates the final output list of significant GO
    terms for each gene in the input list with genes of interest from an
    experiment, for example differentially expressed genes or CRISPR screen
    hits.
    If an input gene is not in the output file, this could have the following
    reasons:
    1) No corresponding HGNC gene symbol, HGNC:ID and/or UniProt:ID could be
    identified (for all --id_type values except custom). All are required to map 
    human genes and assemble their GO annotations. See the genewalk log 
    file for all genes filtered out this way.
    
    2) (if alpha_FDR set to < 1) no GO terms were significant at the
    chosen significance level alpha_FDR.
    
    3) (in case of mouse or rat genes) no mapped human ortholog was identified.

    See the genewalk log file to see all genes filtered out because of 1) or 3).
    
    If a gene is listed in the output file with a value >= 0 in the column 
    ncon_gene but without any listed GO annotations: no GO annotations (with the
    right GO evidence codes) could be retrieved. If a gene is listed but has no
    ncon_gene value (NaN) in the output file: the gene was correctly mapped but
    could not be included in the GeneWalk network. This scenario is uncommon and
    warrants further inspection.

    Parameters
    ----------
    graph : networkx.MultiGraph
        GeneWalk network for which the statistics are calculated.
    genes : list of dict
        List of gene references for relevant genes.
    nvs : list of dict
        Node vectors for nodes in the graph.
    null_dist : np.array
        Similarity random (null) distribution.
    gene_id_type : Optional[str]
        The type of gene IDs that were the basis of doing the analysis.
        In case of mgi_id, rgd_id or ensembl_id, we prepend a column to
        the table for MGI, RGD, or ENSEMBL IDs, respectively. If custom,
        the input genes were not mapped to human genes, so the hgnc_symbol
        and hgnc_id columns are not present in the output.
        Default: hgnc_symbol
    """
    def __init__(self, graph, genes, nvs, null_dist,
                 gene_id_type='hgnc_symbol'):
        self.graph = graph
        self.genes = genes
        self.nvs = nvs
        self.srd = null_dist
        self.gene_id_type = gene_id_type
        self.go_nodes = set(nx.get_node_attributes(self.graph, 'GO'))
        self.gene_nodes = self.get_gene_nodes()

    def get_gene_nodes(self):
        if self.gene_id_type == 'custom':
            return {g['ID'] for g in self.genes}
        else:
            return {g['HGNC_SYMBOL'] for g in self.genes}

    def get_gene_node_id(self, gene):
        return gene['HGNC_SYMBOL'] if self.gene_id_type != 'custom' \
            else gene['ID']

    def get_gene_attribs(self, gene):
        """Return an attribute dict for a given gene."""
        gene_id = self.get_gene_node_id(gene)
        if gene_id in self.graph:
            ncon_gene = len(self.graph[gene_id])
        else:
            ncon_gene = np.nan
        hgnc_symbol = gene['HGNC_SYMBOL'] if self.gene_id_type != 'custom' \
            else None
        hgnc_id = gene['HGNC'] if self.gene_id_type != 'custom' else None
        custom_id = gene['ID'] if self.gene_id_type == 'custom' else None
        return {
            'hgnc_symbol': hgnc_symbol,
            'hgnc_id': hgnc_id,
            'custom_id': custom_id,
            'ncon_gene': ncon_gene
        }

    def get_go_attribs(self, gene_attribs, nv, alpha_fdr):
        """Return GO entries and their attributes for a given gene."""
        gene_node_id = gene_attribs['hgnc_symbol'] \
            if self.gene_id_type != 'custom' else gene_attribs['custom_id']
        connected = set(self.graph[gene_node_id]) & self.go_nodes
        go_attribs = []
        pvals = []
        for go_node_id in connected:
            go_attrib = {}
            sim_score = nv.similarity(gene_node_id, go_node_id)
            go_attrib['sim_score'] = sim_score
            go_attrib['go_id'] = go_node_id
            go_attrib['ncon_go'] = len(self.graph[go_node_id])
            go_attrib['go_name'] = self.graph.nodes[go_node_id]['name']
            go_attrib['go_domain'] = \
                self.graph.nodes[go_node_id]['domain'].replace('_', ' ')
            go_attrib['pval'] = self.psim(sim_score)
            pvals.append(go_attrib['pval'])
            go_attribs.append(go_attrib)
        _, qvals = fdrcorrection(pvals, alpha=alpha_fdr, method='indep')
        for idx in range(len(go_attribs)):
            go_attribs[idx]['qval'] = qvals[idx]
        return go_attribs

    def log_stats(self, vals):
        eps = 1e-16
        nreps = len(vals)
        vals = np.asarray(vals)+eps
        g_mean = gmean(vals)-eps
        g_std = gstd(vals) if len(vals) > 1 else eps
        return g_mean, g_mean*(g_std**(-1.96/np.sqrt(nreps))), \
            g_mean*(g_std**(1.96/np.sqrt(nreps)))

    def add_empty_row(self, gene, gene_attribs):
        row = [gene_attribs['hgnc_symbol'],
               gene_attribs['hgnc_id'],
               '', '', '',
               gene_attribs['ncon_gene'],
               np.nan, np.nan, np.nan, np.nan,
               np.nan, np.nan, np.nan, np.nan, np.nan]
        row.extend([np.nan for i in range(len(self.nvs))])
        if self.gene_id_type == 'mgi_id':
            row = [gene.get('MGI', '')] + row
        elif self.gene_id_type == 'rgd_id':
            row = [gene.get('RGD', '')] + row
        elif self.gene_id_type == 'ensembl_id':
            row = [gene.get('ENSEMBL', '')] + row
        elif self.gene_id_type in {'entrez_mouse', 'entrez_human'}:
            row = [gene.get('EGID', '')] + row
        elif self.gene_id_type == 'custom':
            row = [gene.get('ID', '')] + row
        return row

    def generate_output(self, alpha_fdr=1):
        """Main function of GeneWalk object that generates the final
        GeneWalk output table (in csv format).

        Parameters
        ----------
        alpha_fdr : Optional[float]
            Significance level for FDR [0,1] (default=1, i.e. all GO
            terms and their statistics are output). If set to a lower value,
            only connected GO terms with mean padj < alpha_FDR are output.
        """
        rows = []
        for gene in self.genes:
            gene_attribs = self.get_gene_attribs(gene)
            # gene present in GW network
            if not np.isnan(gene_attribs['ncon_gene']):
                all_go_attribs = [self.get_go_attribs(gene_attribs, nv,
                                                      alpha_fdr)
                                  for nv in self.nvs]
                if any(all_go_attribs):  # gene has GO connections
                    go_attrib_dict = {}
                    for go_attrib_list in all_go_attribs:
                        for go_attribs in go_attrib_list:
                            go_id = go_attribs['go_id']
                            if go_id in go_attrib_dict:
                                go_attrib_dict[go_id].append(go_attribs)
                            else:
                                go_attrib_dict[go_id] = [go_attribs]

                    for go_id, go_attribs in go_attrib_dict.items():
                        gene_padj, low_gene_padj, upp_gene_padj = \
                            self.log_stats([attr['qval']
                                            for attr in go_attribs])
                        mean_pval, low_pval, upp_pval = \
                            self.log_stats([attr['pval']
                                            for attr in go_attribs])
                        mean_sim = np.mean([attr['sim_score']
                                            for attr in go_attribs])
                        sem_sim = (np.std([attr['sim_score']
                                           for attr in go_attribs]) /
                                   np.sqrt(len(self.nvs)))
                        if gene_padj < alpha_fdr or alpha_fdr == 1:
                            row = [gene_attribs['hgnc_symbol'],
                                   gene_attribs['hgnc_id'],
                                   go_attribs[0]['go_name'],
                                   go_attribs[0]['go_id'],
                                   go_attribs[0]['go_domain'],
                                   gene_attribs['ncon_gene'],
                                   go_attribs[0]['ncon_go'],
                                   gene_padj, mean_pval, 
                                   mean_sim, sem_sim,
                                   low_gene_padj, upp_gene_padj,
                                   low_pval, upp_pval]
                            row.extend([attr['pval'] for attr in go_attribs])
                            # If dealing with mouse genes, prepend the MGI ID
                            if self.gene_id_type == 'mgi_id':
                                row = [gene.get('MGI', '')] + row
                            # If dealing with rat genes, prepend the RGD ID
                            elif self.gene_id_type == 'rgd_id':
                                row = [gene.get('RGD', '')] + row
                            elif self.gene_id_type == 'ensembl_id':
                                row = [gene.get('ENSEMBL', '')] + row
                            elif self.gene_id_type in \
                                    {'entrez_mouse', 'entrez_human'}:
                                row = [gene.get('EGID', '')] + row
                            elif self.gene_id_type == 'custom':
                                row = [gene.get('ID', '')] + row
                            rows.append(row)
                elif alpha_fdr == 1:  #case: no GO connections
                    row = self.add_empty_row(gene, gene_attribs)
                    rows.append(row)
            elif alpha_fdr == 1:  # case: not in graph
                row = self.add_empty_row(gene, gene_attribs)
                rows.append(row)
        header = ['hgnc_symbol', 'hgnc_id',
                  'go_name', 'go_id', 'go_domain',
                  'ncon_gene', 'ncon_go',
                  'gene_padj', 'pval', 
                  'sim',  'sem_sim',
                  'cilow_gene_padj', 'ciupp_gene_padj',
                  'cilow_pval', 'ciupp_pval']
        header.extend(['pval_rep'+str(i) for i in range(len(self.nvs))])
        if self.gene_id_type in {'mgi_id', 'rgd_id', 'ensembl_id',
                                 'entrez_human', 'entrez_mouse', 'custom'}:
            header = [self.gene_id_type] + header

        df = pd.DataFrame.from_records(rows, columns=header)
        df = self.global_fdr(df,alpha_fdr)
        df.drop(['pval_rep'+str(i) for i in range(len(self.nvs))],
                axis=1, inplace=True) 
        df[self.gene_id_type] = pd.Categorical(df[self.gene_id_type], \
            categories=df[self.gene_id_type].unique(), ordered=True)
        df[['ncon_gene', 'ncon_go']] = \
            df[['ncon_gene', 'ncon_go']].astype('str')
        df = df.sort_values(by=[self.gene_id_type, 'global_padj', 'gene_padj',
                                'sim', 'go_domain', 'go_name'],
                            ascending=[True, True, True, False, True, True])
        if self.gene_id_type == 'custom':
            df.drop(['hgnc_symbol','hgnc_id'],axis=1, inplace=True)
        return df

    def psim(self, sim):
        """Determine the p-value of the experimental similarity by determining
        its percentile, i.e. the normalized rank, in the null distribution
        with random similarity values.
        """
        rank = np.searchsorted(self.srd, sim)
        pct_rank = float(rank) / len(self.srd)
        pval = 1 - pct_rank
        eps = 1e-16
        if pval < eps:
            pval = eps
        return pval

    def global_fdr(self,df,alpha_fdr):
        """Determine the global_padj values through FDR multiple testing 
        correction over all gene - GO annotation pairs present in the output
        file.
        """
        global_stats = {'global_padj': [], 'cilow_global_padj': [],
                        'ciupp_global_padj': []}
        colloc = {'global_padj': 8, 'cilow_global_padj': 4,
                          'ciupp_global_padj': 4} 
        ids = df[~df['pval_rep0'].isna()].index
        qvals = np.empty((len(ids),len(self.nvs)))
        qvals[:] = np.nan
        for i in range(len(self.nvs)):
            _, qvals[:,i] = fdrcorrection(df['pval_rep'+str(i)][ids], 
                                        alpha=alpha_fdr, method='indep')
        for i in range(qvals.shape[0]):
            mean_padj, low_padj, upp_padj = self.log_stats(qvals[i,:])
            global_stats['global_padj'].append(mean_padj)
            global_stats['cilow_global_padj'].append(low_padj)
            global_stats['ciupp_global_padj'].append(upp_padj)
        for key in global_stats.keys():
            df.insert((len(df.columns)-colloc[key]-len(self.nvs)),key,np.nan)
            df.loc[ids,key] = global_stats[key]
        return df
